"""s3-delta target sink class, which handles writing streams."""

from __future__ import annotations

import os
from typing import Dict, Optional

from singer_sdk import PluginBase
from singer_sdk.helpers._typing import DatetimeErrorTreatmentEnum
from singer_sdk.sinks import BatchSink
import pandas as pd

from target_s3_delta.common import ExtractMode
from target_s3_delta.utils import float_to_decimal

TEMP_DATA_DIRECTORY = "/tmp/meltano_temp_data/"
MAX_SIZE_DEFAULT = 200000


class S3DeltaSink(BatchSink):
    """s3-delta target sink class."""

    @property
    def datetime_error_treatment(self) -> DatetimeErrorTreatmentEnum:
        return DatetimeErrorTreatmentEnum.NULL

    @property
    def mode(self) -> ExtractMode:
        """Get extract mode"""
        return self.config.get("mode")

    @property
    def max_size(self) -> int:
        """Get max batch size.

        Returns:
            Max number of records to batch before `is_full=True`
        """
        return int(self.config.get("batch_size", MAX_SIZE_DEFAULT))

    def _validate_and_parse(self, record: dict) -> dict:
        """Validate or repair the record, parsing to python-native types as needed."""
        record = float_to_decimal(record)
        return super()._validate_and_parse(record=record)

    def process_record(self, record: dict, context: dict) -> None:
        """Load the latest record from the stream.
        Args:
            record: Individual record in the stream.
            context: Stream partition or context dictionary.
        """
        if "records" not in context:
            context["records"] = []

        context["records"].append(record)

    def start_batch(self, context: dict) -> None:
        """Start a batch.

        Developers may optionally add additional markers to the `context` dict,
        which is unique to this batch.

        Args:
            context: Stream partition or context dictionary.
        """
        batch_key = context["batch_id"]
        context["file_path"] = f"{batch_key}.parquet"

    def process_batch(self, context: dict) -> None:
        """Write out any prepped records and return once fully written.

        Args:
            context: Stream partition or context dictionary.
        """
        self.logger.info("Batch processing has started.")
        is_exist = os.path.exists(TEMP_DATA_DIRECTORY)
        if not is_exist:
            os.makedirs(TEMP_DATA_DIRECTORY)

        df = pd.DataFrame(context["records"])
        df.to_parquet(f"{TEMP_DATA_DIRECTORY}{context.get('file_path')}", engine="pyarrow")

        context["records"] = []
        self.logger.info("Batch processing has finished.")



import decimal
import math
from decimal import Decimal


def float_to_decimal(value):
    """
    Walk the given data structure and turn all instances of float into double.
    """
    if isinstance(value, float):
        return Decimal(str(value))
    if isinstance(value, list):
        return [float_to_decimal(child) for child in value]
    if isinstance(value, dict):
        return {k: float_to_decimal(v) for k, v in value.items()}
    return value


def numeric_schema_with_precision(schema):
    if "type" not in schema:
        return False
    if isinstance(schema["type"], list):
        if "number" not in schema["type"]:
            return False
    elif schema["type"] != "number":
        return False
    if "multipleOf" in schema:
        return True
    return "minimum" in schema or "maximum" in schema


def walk_schema_for_numeric_precision(schema):
    # Added a default max precision for cases when the schema does not specify a maximum or minimum precision
    # Set this default based on https://tools.ietf.org/html/rfc7159#section-6
    default_maximum_value = 9007199254740991
    if isinstance(schema, list):
        for v in schema:
            walk_schema_for_numeric_precision(v)
    elif isinstance(schema, dict):
        if numeric_schema_with_precision(schema):

            def get_precision(key, default=1):
                v = abs(Decimal(schema.get(key, default))).log10()
                if v < 0:
                    return round(math.floor(v))
                return round(math.ceil(v))

            scale = -1 * get_precision("multipleOf")
            digits = max(
                get_precision("minimum", default_maximum_value), get_precision("maximum", default_maximum_value)
            )
            precision = digits + scale
            if decimal.getcontext().prec < precision:
                decimal.getcontext().prec = precision
        else:
            for v in schema.values():
                walk_schema_for_numeric_precision(v)
